#!/usr/bin/perl
#
# file: wsj_exp/wsj_run.pl.in
#
$ISIP_HELP_FILE = <<__ISIP_HELP_HERE_FILE__;
name: wsj_run
synopsis: 
    wsj_run -num_features <number> -mode <mode> -feature_format <format> 
    -mixtures <num_of_mixtures_for_xwrd_models>
    -train_beam_pruning <beam_pruning_for_training>
    -split_threshold <threshold_to_split_a_node_in_a_decision_tree>
    -merge_threshold <threshold_to_merge_two_nodes_in_a_decision_tree>
    -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_a_decision_tree>
    -lm <language_model> -lm_scale <language_model_scale> 
    -word_penalty <word_insertion_penalty> 
    -test_state_beam_pruning <state_level_beam_pruning_for_testing>
    -test_model_beam_pruning <model_level_beam_pruning_for_testing>
    -test_word_beam_pruning <word_level_beam_pruning_for_testing>
    -models_path <base_directory_of_models>
    -train_mfc_list <list_of_training_feature_files> 
    -test_mfc_list <list_of_testing_feature_files>
    -cpus_train <cpus_used_for_training>
    -cpus_test <cpus_used_for_testing>
    -cluster_train <cluster_used_for_training>
    -cluster_test <cluster_used_for_testing>
    -num_nodes <number_of_cluster_nodes>
    -walltime <expected_walltime_on_the_cluster>
    
descr: 
    This program runs a complete experiment on WSJ data.  The user user
    may set all relevant parameters for the recognizer, such as the
    number and type of features.  The default values for all parameters
    are the values used in the baseline system on ISIP frontend except for 
    the test beams that have been set to 1000 according to the ETSI frontend.
    Training and/or decoding can be run using these scripts.

example:
 
1) To compute delta and acceleration coefficients on the fly and use 
   two cpus on isip204 & one cpu on isip105:
   wsj_run -num_features 13 -mode acc -feature_format htk -mixtures 1 
    -lm devel -train_mfc_list train.list -test_mfc_list test.list
    -cpus_train isip204 isip204 isip105 -cpus_test isip204 isip205 isip206

2) Do not compute deltas or accelerations on the fly and use two cpus
   on isip204 & one cpu on isip105:
   wsj_run -num_features 39 -feature_format htk -mixtures 1 -lm devel
   -train_mfc_list train.list -test_mfc_list test.list
    -cpus_train isip204 isip105 -cpus_test isip204 isip205 isip206

3) To run only the training:
   cd exp_070; 
   wsj_run -num_features 13 -mode acc -feature_format htk -mixtures 1 
    -lm devel -train_mfc_list train.list -cpus_train isip204 isip204 isip105  

4) To run only decoding using models trained in exp_070 directory.
   cd exp_071; 
   wsj_run -num_features 13 -mode acc -feature_format htk -lm devel
    -test_mfc_list test.list -cpus_test isip214 isip215 isip216
    -models_path /isip/data/exp/wsj/exp_070

5) To run only the training on cluster:
   cd exp_070; 
   wsj_run -num_features 13 -mode acc -feature_format htk -mixtures 1 
    -lm devel -train_mfc_list train.list -cluster_train maverick -num_nodes 30

6) To run only decoding using models trained on cluster in exp_070 directory.
   cd exp_071; 
   wsj_run -num_features 13 -mode acc -feature_format htk -lm devel
    -test_mfc_list test.list -cluster_test maverick -num_nodes 30
    -models_path /isip/data/exp/wsj/exp_070

options: 

   -num_features <number> : 
      number of features per frame in each of the input data files
      (default is 39)

   -mode <mode> : 
      this option tells the script whether you want to compute deltas or both 
      delta/accelerations or neither on the fly. Choices are:
      delta    : computes deltas on the fly
      acc      : computes accelerations on the fly
      none     : does not computes deltas or accelerations on the fly (default)

   -feature_format <format> : 
      specifies the feature format to use. Choices are:
      htk        : HTK format features (default)
      isip_proto : ISIP format features

   -mixtures <num_of_mixtures_for_xwrd_models> : 
      specifies the number od cross-word mixtures to train and use for 
      decoding. Choices are:
       1 (default)
       2
       4
       8
      16

   -train_beam_pruning <beam_pruning_for_training> : 
      specifies the beam pruning for all the three levels during training
      (default is 1000)

   -split_threshold <threshold_to_split_a_node_in_decision_tree> : 
      specifies the minimum increase in likelihood to split a node in the 
      decision tree during the state tying process (default is 165)

   -merge_threshold <threshold_to_merge_two_node_in_decision_tree> : 
      specifies the maximum decrease in likelihood to merge two nodes in the
      decision tree during the state tying process (default is 165)

   -num_occ_threshold <min_num_of_occupancy_to_split_a_node_in_decision_tree> :
      specifies the minimum number of states for splitting a node in the 
      decision tree during the state tying process (default is 840)

   -lm <language_model> : 
      specifies the language model for testing. Choices are:
      devel : language model for WSJ devtest set (default)
      eval  : language model for WSJ evaltest set

   -lm_scale <language_model_scale> : 
      specifies the weight for the langauge model score (default is 18)

   -word_penalty <word_insertion_penalty> : 
      specifies the penalty applied to each word that is hypothesized 
      (default is 10)

   -test_state_beam_pruning <state_level_beam_pruning_for_testing> : 
      specifies the state-level beam pruning during decoding (default is 300) 

   -test_model_beam_pruning <model_level_beam_pruning_for_testing> : 
      specifies the model-level beam pruning during decoding (default is 250) 

   -test_word_beam_pruning <word_level_beam_pruning_for_testing> : 
      specifies the word-level beam pruning during decoding (default is 250) 

   -train_mfc_list <list_of_training_feature_files> : 
      specifies the training feature files list

   -test_mfc_list <list_of_test_feature_files> :
      specifies the testing feature files list
 
   -cpus_train <cpus_used_to_train>
      specifies the cpus to be employed for training

   -cpus_test <cpus_used_to_test>
      specifies the cpus to be employed for testing
    
    -cluster_train <cluster_used_to_train>
    specifies the cluster to be employed for training
    
    -cluster_test <cluster_used_to_test>
    specifies the cluster to be employed for testing
    
    -num_nodes <number_of_cluster_nodes>
    specifies the number of nodes required to run jobs
    
    -walltime <expected_walltime_on_the_cluster>
    specifies the walltime expected for the jobs to finish on the cluster
    
    -models_path <base_directory_of_models>
    specifies the models base-directory to be employed for testing
      (default is the current directory)

   -clean :  
      if this flag is thrown then all previous experiments run will be cleaned
      up (i.e. deleted). All the intermediate and output files in the 
      experiment directory will be deleted. This flag is used to start from a
      fresh state.

arguments: none
 
man page: none
__ISIP_HELP_HERE_FILE__

# define the path to this script
#
$WSJ_PATH = "/cavs/hse/ies/isip/users/raghavan/aurora/cluster_enabled/asr_wsj_tutorial_v1.4.2";

# system packages
#
use File::Basename;
use File::Path;
use File::Copy;
use FileHandle;
autoflush STDOUT;

# isip libraries
#
require "$WSJ_PATH/lib/scripts/perl/command_line.pm";
require "$WSJ_PATH/lib/scripts/perl/wsj_subs.pm";

#----------------------------------------------------------------------------
#
# constant and global variable definitions
#
#----------------------------------------------------------------------------

# define the program name
#
@tmp = split(/\//, $0);
$ISIP_PROG = pop(@tmp);

# define a separator for output comments
#
$TI_LINE_SEP = "-------------------  $ISIP_PROG  ----------------";

# define files and locations to be used by this script
#
$TI_PATH_TO_TRAIN = "$WSJ_PATH/data/train";
$TI_PATH_TO_OUTPUT = "$WSJ_PATH/exp/train";
$TI_PATH_TO_BW = "$TI_PATH_TO_OUTPUT/baum_welch";
$TI_PATH_TO_VIT = "$TI_PATH_TO_OUTPUT/viterbi";
$TI_PATH_TO_MFC = "$WSJ_PATH/data/data_generation";
$TI_TRAIN_MFC_FILE_LIST = "$TI_PATH_TO_MFC/lists/train_mfc.list";

# define the sub-scripts called by this script
#
$TI_PATH_TO_SCRIPTS = "$WSJ_PATH/bin/scripts";
#$TI_PREPROCESS_SCRIPT = "$TI_PATH_TO_SCRIPTS/wsj_preprocess";
$TI_TRAINING_SCRIPT = "$TI_PATH_TO_SCRIPTS/wsj_train";
$TI_DECODING_SCRIPT = "$TI_PATH_TO_SCRIPTS/wsj_decode";
$TI_PATH_TO_EXP = "$WSJ_PATH/exp";

# define command-line options
#
$TI_OPT_NUM_FEATURES = "-num_features";
$TI_OPT_MODE = "-mode";
$TI_OPT_FEAT_FORMAT = "-feature_format";
$TI_OPT_NUM_MIX = "-mixtures";
$TI_OPT_TR_BEAM_PRUN = "-train_beam_pruning";
$TI_OPT_SP_THRESH = "-split_threshold";
$TI_OPT_ME_THRESH = "-merge_threshold";
$TI_OPT_NUM_OCC_THRESH = "-num_occ_threshold";
$TI_OPT_LM = "-lm";
$TI_OPT_LM_SCALE = "-lm_scale";
$TI_OPT_WORD_PEN = "-word_penalty";
$TI_OPT_TE_ST_BEAM_PRUN = "-test_state_beam_pruning";
$TI_OPT_TE_MO_BEAM_PRUN = "-test_model_beam_pruning";
$TI_OPT_TE_WORD_BEAM_PRUN = "-test_word_beam_pruning";
$TI_OPT_TRAIN_MFC_LIST = "-train_mfc_list";
$TI_OPT_TEST_MFC_LIST = "-test_mfc_list";
$TI_OPT_CPUS_TRAIN = "-cpus_train";
$TI_OPT_CPUS_TEST = "-cpus_test";
$TI_OPT_CLUSTER_TRAIN = "-cluster_train";
$TI_OPT_CLUSTER_TEST = "-cluster_test";
$TI_OPT_NUM_NODES = "-num_nodes";
$TI_OPT_WALLTIME = "-walltime";
$TI_OPT_MODELS_PATH = "-models_path";
$TI_OPT_CLEAN = "-clean";

$TI_MODE_DELTA = "delta";
$TI_MODE_ACC = "acc";
$TI_MODE_NONE = "none";
@TI_MODE_TYPES = ($TI_MODE_DELTA, $TI_MODE_ACC, $TI_MODE_NONE);

$TI_FEAT_FORMAT_HTK = "htk";
$TI_FEAT_FORMAT_ISIP = "isip_proto";
@TI_FEAT_FORMAT_TYPES = ($TI_FEAT_FORMAT_HTK, $TI_FEAT_FORMAT_ISIP);

$TI_NUM_MIX_1 = "1";
$TI_NUM_MIX_2 = "2";
$TI_NUM_MIX_4 = "4";
$TI_NUM_MIX_8 = "8";
$TI_NUM_MIX_16 = "16";
@TI_NUM_MIX_TYPES = ($TI_NUM_MIX_1, $TI_NUM_MIX_2, $TI_NUM_MIX_4, 
		     $TI_NUM_MIX_8, $TI_NUM_MIX_16);

$TI_LM_DEVEL = "devel";
$TI_LM_EVAL = "eval";
@TI_LM_TYPES = ($TI_LM_DEVEL, $TI_LM_EVAL);

#define the defaults
#
$TI_NUM_FEATURES_DEF = "39";
$TI_MODE_DEF = $TI_MODE_NONE;
$TI_FEAT_FORMAT_DEF = $TI_FEAT_FORMAT_HTK;
$TI_NUM_MIX_DEF = $TI_NUM_MIX_1;
$TI_TR_BEAM_PRUN_DEF = "1000";
$TI_SP_THRESH_DEF = "165";
$TI_ME_THRESH_DEF = "165";
$TI_NUM_OCC_THRESH_DEF = "840";
$TI_LM_DEF = $TI_LM_DEVEL;
$TI_LM_SCALE_DEF = "18";
$TI_WORD_PEN_DEF = "10";
$TI_TE_ST_BEAM_PRUN_DEF = "300";
$TI_TE_MO_BEAM_PRUN_DEF = "250";
$TI_TE_WORD_BEAM_PRUN_DEF = "250";

# define the local host
#
print "INFO: wsj_run - checking the localhost name\n";
$local_host = `hostname`;
print "INFO: wsj_run - the localhost is $local_host";

# extract the machine name from the local_host
#
@name = split(/\./, $local_host);
$host_name = $name[0];

#----------------------------------------------------------------------------
#
# command line processing
#
#----------------------------------------------------------------------------
# print a debugging message
#
print "$TI_LINE_SEP\n";
print "INFO: wsj_run - checking command-line arguments\n";

# parse the command line
#
my ($num_features, $mode, $feature_format, $mixtures, $train_beam_pruning, 
    $split_threshold, $merge_threshold, $num_occ_threshold, $lm, $lm_scale,
    $word_penalty, $test_state_beam_pruning, $test_model_beam_pruning,
    $test_word_beam_pruning, $train_mfc_list, $test_mfc_list, $cpus_train,
    $cpus_test, $cluster_train, $cluster_test, $num_nodes, $walltime, $models_path, $clean) =  
    command_line(0, $TI_OPT_NUM_FEATURES, 1, $TI_OPT_MODE, 1,
		 $TI_OPT_FEAT_FORMAT, 1, $TI_OPT_NUM_MIX, 1,
		 $TI_OPT_TR_BEAM_PRUN, 1,
		 $TI_OPT_SP_THRESH, 1, $TI_OPT_ME_THRESH, 1,
		 $TI_OPT_NUM_OCC_THRESH, 1, $TI_OPT_LM, 1, 
		 $TI_OPT_LM_SCALE, 1, $TI_OPT_WORD_PEN, 1,
		 $TI_OPT_TE_ST_BEAM_PRUN, 1, 
		 $TI_OPT_TE_MO_BEAM_PRUN, 1, 
		 $TI_OPT_TE_WORD_BEAM_PRUN, 1,
		 $TI_OPT_TRAIN_MFC_LIST, 1, 
		 $TI_OPT_TEST_MFC_LIST, 1,
		 $TI_OPT_CPUS_TRAIN, -1, $TI_OPT_CPUS_TEST, -1,
		 $TI_OPT_CLUSTER_TRAIN, 1,$TI_OPT_CLUSTER_TEST, 1,
		 $TI_OPT_NUM_NODES, 1,
		 $TI_OPT_WALLTIME, 1,
		 $TI_OPT_MODELS_PATH, 1, $TI_OPT_CLEAN, 0);

# set defaults if necessary
#
# clean up if desired
#
if ($clean ne "") {
    (run_cleanup() == 1) or isip_die("ERROR: cleanup failed...exiting");
    print "INFO: wsj_run - cleanup succeeded...exiting\n";
    exit 0;
}

if ($num_features eq "") {
    $num_features = $TI_NUM_FEATURES_DEF;
}

if ($mode eq "") {
    $mode = $TI_MODE_DEF;
}

if ($feature_format eq "") {
    $feature_format = $TI_FEAT_FORMAT_DEF;
}

if ($mixtures eq "") {
    $mixtures = $TI_NUM_MIX_DEF;
}

if ($train_beam_pruning eq "") {
    $train_beam_pruning = $TI_TR_BEAM_PRUN_DEF;
}

if ($split_threshold eq "") {
    $split_threshold = $TI_SP_THRESH_DEF;
}

if ($merge_threshold eq "") {
    $merge_threshold = $TI_ME_THRESH_DEF;
}

if ($num_occ_threshold eq "") {
    $num_occ_threshold = $TI_NUM_OCC_THRESH_DEF;
}

if ($lm eq "") {
    $lm = $TI_LM_DEF;
}

if ($lm_scale eq "") {
    $lm_scale = $TI_LM_SCALE_DEF;
}

if ($word_penalty eq "") {
    $word_penalty = $TI_WORD_PEN_DEF;
}

if ($test_state_beam_pruning eq "") {
    $test_state_beam_pruning = $TI_TE_ST_BEAM_PRUN_DEF;
}

if ($test_model_beam_pruning eq "") {
    $test_model_beam_pruning = $TI_TE_MO_BEAM_PRUN_DEF;
}

if ($test_word_beam_pruning eq "") {
    $test_word_beam_pruning = $TI_TE_WORD_BEAM_PRUN_DEF;
}
if ($models_path eq  "") {
    $models_path = "$WSJ_PATH/exp";
}

# use the default host cpu only if no training or testing cps are given from 
# the commandline and the training/testing mfc lists are specified 
# respectively 
#
if (($cpus_train eq "") && ($train_mfc_list ne "")) {
    $cpus_train = $host_name;
}

if (($cpus_test eq "") && ($test_mfc_list ne "")) {
    $cpus_test = $host_name;
}

# verify mode, feature_format and langauage model type are valid
#
$mode_v = grep /\b$mode\b/, @TI_MODE_TYPES;
$feature_format_v = grep /\b$feature_format\b/, @TI_FEAT_FORMAT_TYPES;
$mixtures_v = grep /\b$mixtures\b/, @TI_NUM_MIX_TYPES;
$lm_v = grep /\b$lm\b/, @TI_LM_TYPES;

if ($mode_v < 1) {
    $valid_modes = join(", ", @TI_MODE_TYPES);
    isip_die ("ERROR: $mode_type is not a valid mode type\n ERROR: valid modes are $valid_modes");
}

if ($feature_format_v < 1) {
    $valid_formats = join(", ", @TI_FEAT_FORMAT_TYPES);
    isip_die ("ERROR: $feature_format is not a valid feature format\nERROR: valid feature formats are $valid_formats");
}

if ($mixtures_v < 1) {
    $valid_mixtures = join(", ", @TI_NUM_MIX_TYPES);
    isip_die ("ERROR: $mixtures is not a valid number of mixtures\nERROR: valid number of mixtures are $valid_mixtures");
}

if ($lm_v < 1) {
    $valid_types = join(", ", @TI_LM_TYPES);
    isip_die ("ERROR: $lm is not a valid language model\nERROR: valid language types are $valid_types");
}

print "INFO: wsj_run - running with parameters: \n";
$out_line = join (" ", $TI_OPT_NUM_FEATURES, $num_features, $TI_OPT_MODE,
		  $mode, $TI_OPT_FEAT_FORMAT, $feature_format,
		  $TI_OPT_NUM_MIX, $mixtures, 
		  $TI_OPT_TR_BEAM_PRUN, $train_beam_pruning, 
		  $TI_OPT_SP_THRESH, $split_threshold, 
		  $TI_OPT_ME_THRESH, $merge_threshold, 
		  $TI_OPT_NUM_OCC_THRESH, $num_occ_threshold, 
		  $TI_OPT_LM, $lm, $TI_OPT_LM_SCALE, $lm_scale, 
		  $TI_OPT_WORD_PEN, $word_penalty, 
		  $TI_OPT_TE_ST_BEAM_PRUN, $test_state_beam_pruning, 
		  $TI_OPT_TE_MO__BEAM_PRUN, $test_model_beam_pruning, 
		  $TI_OPT_TE_WORD_BEAM_PRUN, $test_word_beam_pruning, 
		  $TI_OPT_TRAIN_MFC_LIST, $train_mfc_list, 
		  $TI_OPT_TEST_MFC_LIST, $test_mfc_list,
		  $TI_OPT_CPUS_TRAIN, $cpus_train, 
		  $TI_OPT_CPUS_TEST, $cpus_test, 
		  $TI_OPT_CLUSTER_TRAIN, $cluster_train, 
		  $TI_OPT_CLUSTER_TEST, $cluster_test,
		  $TI_OPT_NUM_NODES, $num_nodes,
		  $TI_OPT_WALLTIME, $walltime,
		  $TI_OPT_MODELS_PATH, $models_path);

print "$out_line\n";

#----------------------------------------------------------------------------
#
# generate training and testing mfcc lists and corresponding training 
# transcriptions for jobs on cpu's
#
#----------------------------------------------------------------------------

#print "INFO: wsj_run - generating training and testing mfcc lists and\n";
#print "INFO: wsj_run - corresponding training transcriptions for jobs on the cpu's\n";

# generate training lists, corresponding transcriptions and test lists
#
#@args = ($TI_PREPROCESS_SCRIPT, $TI_OPT_TRAIN_MFC_LIST, $train_mfc_list, 
	# $TI_OPT_TEST_MFC_LIST, $test_mfc_list, $TI_OPT_CPUS_TRAIN, 
	# $cpus_train, $TI_OPT_CPUS_TEST, $cpus_test);

#$out_line = join(" ", @args);
#print "COMMAND: $out_line\n";

#(system(@args) == 0) 
#    or isip_die ("ERROR: failed call to $TI_PREPROCESS_SCRIPT\n ERROR: unable to generate file transcriptions...exiting");

#print "$TI_LINE_SEP\n";

#----------------------------------------------------------------------------
#
# train the models if required
#
#----------------------------------------------------------------------------

if ($train_mfc_list ne "") {
    @args = ($TI_TRAINING_SCRIPT, $TI_OPT_NUM_FEATURES, $num_features, 
	     $TI_OPT_MODE, $mode, $TI_OPT_FEAT_FORMAT, $feature_format,
	     $TI_OPT_NUM_MIX, $mixtures,
	     $TI_OPT_TR_BEAM_PRUN, $train_beam_pruning, 
	     $TI_OPT_SP_THRESH, $split_threshold, 
	     $TI_OPT_ME_THRESH, $merge_threshold, 
	     $TI_OPT_NUM_OCC_THRESH, $num_occ_threshold, 
	     $TI_OPT_TRAIN_MFC_LIST, $train_mfc_list,
	     $TI_OPT_CPUS_TRAIN, $cpus_train,
	     $TI_OPT_CLUSTER_TRAIN, $cluster_train, 
	     $TI_OPT_NUM_NODES, $num_nodes, 
	     $TI_OPT_WALLTIME, $walltime);
    
    print "INFO: wsj_run - training models with command-line:\n";
    $out_line = join(" ", @args);
    print "COMMAND: $out_line\n";
    
    # train
    #
    (system(@args) == 0) or
	isip_die ("ERROR: failed call to $TI_TRAINING_SCRIPT\n ERROR: unable to train models...exiting");
    
    print "$TI_LINE_SEP\n";
    print "INFO: wsj_run - finished training\n";
    
}

#----------------------------------------------------------------------------
#
# decode the data using the trained models
#
#----------------------------------------------------------------------------

if ($test_mfc_list ne "") {
    @args = ($TI_DECODING_SCRIPT, $TI_OPT_MODE, $mode, 
	     $TI_OPT_FEAT_FORMAT, $feature_format, $TI_OPT_LM, $lm, 
	     $TI_OPT_LM_SCALE, $lm_scale, $TI_OPT_WORD_PEN, $word_penalty,
	     $TI_OPT_TE_ST_BEAM_PRUN, $test_state_beam_pruning,
	     $TI_OPT_TE_MO_BEAM_PRUN, $test_model_beam_pruning,
	     $TI_OPT_TE_WORD_BEAM_PRUN, $test_word_beam_pruning,
	     $TI_OPT_TEST_MFC_LIST, $test_mfc_list, 
	     $TI_OPT_CPUS_TEST, $cpus_test, $TI_OPT_MODELS_PATH, $models_path);
    
    print "INFO: wsj_run - decoding models with command-line:\n";
    $out_line = join(" ", @args);
    print "COMMAND: $out_line\n";

   # decode
   #
    (system(@args) == 0) or isip_die ("ERROR: failed call to $TI_DECODING_SCRIPT\n ERROR: unable to decode data...exiting");

    print "$TI_LINE_SEP\n";
    print "INFO: wsj_run - finished decoding\n";
}

#----------------------------------------------------------------------------
#
# subroutines
#
#----------------------------------------------------------------------------

# subroutine: run_cleanup
#
# arguments: none
#
# description: this routine cleans up the experiment directory
#
# return: 1 if error, 0 if no error
#
sub run_cleanup() {

    print "INFO: run_cleanup - cleaning up old experiments:\n";

    # run the command
    #
    if (rmtree($TI_PATH_TO_EXP) < 0) {
	print "ERROR: run_cleanup - could not remove $TI_PATH_TO_EXP\n";
	return 0;
    }
    return 1;
}
